package com.xaiten.flutter_x_nfc;

import android.content.Context;
import android.nfc.FormatException;
import android.nfc.tech.NfcA;
import android.os.Handler;
import android.os.Looper;
import android.nfc.NfcAdapter;
import android.nfc.NfcManager;

import android.nfc.Tag;
import android.nfc.tech.Ndef;
import android.os.Build;
import android.Manifest;

import java.io.IOException;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Map;

import io.flutter.Log;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.EventChannel;
import io.flutter.plugin.common.EventChannel.EventSink;
import io.flutter.plugin.common.EventChannel.StreamHandler;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.MethodChannel.MethodCallHandler;
import io.flutter.plugin.common.MethodChannel.Result;
import io.flutter.plugin.common.PluginRegistry.Registrar;

/** FlutterXNfcPlugin */
public class FlutterXNfcPlugin implements MethodCallHandler, StreamHandler, NfcAdapter.ReaderCallback {

  final static String METHOD_NFC_READ = "NfcRead";
  final static String METHOD_NFC_STOP = "NfcStop";

  final static String EVENT_CHANNEL = "flutter_x_nfc/event_channel";
  final static String PLUGIN_TAG = "flutter_x_nfc";

  final static int PERMISSION_NFC = 1007;

  private final static int READER_FLAGS = NfcAdapter.FLAG_READER_NFC_A;

  private NfcAdapter nfcAdapter;
  private Registrar registrar;

  private EventChannel eventChannel;
  private EventSink eventSink;
  private Handler eventChannelHandler;

  public FlutterXNfcPlugin(NfcManager nfcManager, Registrar registrar){
    this.registrar = registrar;
    this.nfcAdapter = nfcManager.getDefaultAdapter();
    this.eventChannel=new EventChannel(registrar.messenger(),EVENT_CHANNEL);
    eventChannel.setStreamHandler(this);
    eventChannelHandler = new Handler(Looper.getMainLooper());
  }

  /** Plugin registration. */
  public static void registerWith(Registrar registrar) {
    if(registrar==null){
      return;
    }

    NfcManager nfcManager = (NfcManager) registrar.activity().getSystemService(Context.NFC_SERVICE);

    if(nfcManager==null){
      return;
    }

    final MethodChannel channel = new MethodChannel(registrar.messenger(), "flutter_x_nfc");

    channel.setMethodCallHandler(new FlutterXNfcPlugin(nfcManager,registrar));
  }

  private static class MethodResultWrapper implements Result {
    private MethodChannel.Result methodResult;
    private Handler handler;

    MethodResultWrapper(MethodChannel.Result result) {
      methodResult = result;
      handler = new Handler(Looper.getMainLooper());
    }

    @Override
    public void success(final Object result) {
      handler.post(
              new Runnable() {
                @Override
                public void run() {
                  methodResult.success(result);
                }
              });
    }

    @Override
    public void error(
            final String errorCode, final String errorMessage, final Object errorDetails) {
      handler.post(
              new Runnable() {
                @Override
                public void run() {
                  methodResult.error(errorCode, errorMessage, errorDetails);
                }
              });
    }

    @Override
    public void notImplemented() {
      handler.post(
              new Runnable() {
                @Override
                public void run() {
                  methodResult.notImplemented();
                }
              });
    }
  }

  @Override
  public void onMethodCall(MethodCall call, Result rawResult) {
    if (registrar.activity() == null) {
      rawResult.error("no_activity", "image_picker plugin requires a foreground activity.", null);
      return;
    }

    MethodResultWrapper result = new MethodResultWrapper(rawResult);

    if (call.method.equals("getPlatformVersion")) {
      result.success("Android " + android.os.Build.VERSION.RELEASE);
    } else if(call.method.equals(METHOD_NFC_READ)){
      startRead(result);
    }else if(call.method.equals(METHOD_NFC_STOP)){
      stopRead(result);
    }
    else {
      result.notImplemented();
    }
  }


  private void startRead(MethodResultWrapper result){

    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
      registrar.activity().requestPermissions(new String[]{Manifest.permission.NFC},PERMISSION_NFC);
    }else{
      result.error("PERMISSION_REQUEST_FAIL","Cannot gaint sufficient permission.",null);
      return;
    }

    if(nfcAdapter!=null && nfcAdapter.isEnabled()){
      if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.KITKAT) {
        nfcAdapter.enableReaderMode(registrar.activity(), this, READER_FLAGS, null );
      }
      result.success(null);
    }else{
      result.error("NFC_FAIL","Cannot get NFC hardware or NFC is disabled",null);
    }
  }

  private void stopRead(MethodResultWrapper result){
    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.KITKAT) {
      nfcAdapter.disableReaderMode(registrar.activity());
    }
    eventSink = null;
    if(result!=null){
      result.success(null);
    }
  }

  private void submitEventChannelData(final Map<String,Object> data){
    eventChannelHandler.post(new Runnable() {
      @Override
      public void run() {
        if(eventSink!=null){
          eventSink.success(data);
        }
      }
    });
  }

  private String bytesToHexString(byte[] src) {
    StringBuilder builder = new StringBuilder("0x");

    if (src == null || src.length>0) {
      return null;
    }

    for (int a=0;a<src.length;a++) {
      builder.append(String.format("%02x", src[a]));
    }

    return builder.toString();
  }



  @Override
  public void onTagDiscovered(Tag tag) {

    Ndef nTag = Ndef.get(tag);
    try{
      if(nTag!=null){
        nTag.connect();
      }

      String message = new String(nTag.getNdefMessage().toByteArray(), Charset.forName("UTF-8"));
      String id = bytesToHexString(tag.getId());

      Map<String,Object> toSubmit = new HashMap<>();
      toSubmit.put("id",id);
      toSubmit.put("message",message);

      submitEventChannelData(toSubmit);

    }catch (IOException e){
      Log.e(PLUGIN_TAG,"Cannot connect to the tag",e);
    } catch (FormatException e) {
      Log.e(PLUGIN_TAG,"Cannot read the tag",e);
    }


  }

  @Override
  public void onListen(Object o, EventChannel.EventSink eventSink) {
    this.eventSink = eventSink;
  }

  @Override
  public void onCancel(Object o) {
    eventSink=null;
    stopRead(null);
  }
}
